
//Game logic.
//Luke Lenhart, 2008
//See the MPMA Framework license for details on how this code may be used.

#include <iostream>

#include "mpma/base/Timer.h"
#include "mpma/input/Keyboard.h"
#include "mpma/base/MiscStuff.h"
#include "mpma/gfxsetup/Setup.h" //to get size for fps output

#include "Render.h"

namespace
{
    //timer and FPS counter
    MPMA::Timer timer;
    uint fpsCount=0;
    float fpsTime=0;

    float gameTime=0;

    //game elements
    float paddleY=0;
    float paddleYResetMagnitude=0;

    GEO::Vector2 ball(0, 0);
    GEO::Vector2 ballDir(0, 0);

    enum GameState
    {
        Starting,
        Playing,
        Ending
    };

    GameState state=Starting;
    float stateTime=0;

    //scene info
    const float paddleX=7;
    const float wallDistance=4.4f;
    const float outOfBoundDistance=8.0f;
}

namespace Game
{
    void Update()
    {
        float time=(float)timer.Step();
        gameTime+=time;

        //output FPS every 5 seconds
        fpsTime+=time;
        ++fpsCount;
        if (fpsTime>=5.0f)
        {
            float actualFPS=fpsCount/fpsTime;
            std::cout<<"FPS (last 5 seconds): "<<actualFPS<<"   Resolution: "<<GFX::GetWindowState()->Width<<"x"<<GFX::GetWindowState()->Height<<std::endl;
            fpsTime=0;
            fpsCount=0;
        }

        //animate light
        Renderer::lightPos.x()=sin(sin(gameTime*0.25f))*10;
        Renderer::lightPos.y()=sin(cos(gameTime*0.333f))*5;
        Renderer::lightPos.z()=-10+sin(gameTime*0.444f)*4;

        //game state
        stateTime+=time;

        if (state==Starting) //setup
        {
            ball=GEO::Vector2(0,0);

            do
            {
                float balldir=rndf()*3.15159f*2;
                ballDir=GEO::Vector2(sin(balldir), cos(balldir));
            } while (ballDir.x()<0.15f && ballDir.x()>-0.15f);
            ballDir*=6.5f+rndf()*0.5f;

            if (stateTime>2)
            {
                Renderer::ball.radius=Renderer::ballSize;

                stateTime=2;
                state=Playing;
            }
            else
            {
                float mod=stateTime*0.5f;
                Renderer::ball.radius=sin(mod*3.141592f*(5.0f/6))*sin(mod*3.141592f*(5.0f/6))*2*(Renderer::ballSize*2);

                paddleY+=paddleYResetMagnitude*time;
            }
        }
        else if (state==Playing) //running
        {
            //controls
            if (time>0.5f) time=0.5f;
            const float speed=5.0f;
            if (INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::UP) || INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::W) || INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::COMMA))
                paddleY+=speed*time;
            if (INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::DOWN) || INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::S) || INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::O))
                paddleY-=speed*time;

            if (paddleY>3.95f) paddleY=3.95f;
            if (paddleY<-3.95f) paddleY=-3.95f;

            //ball update.  rather than doing it correctly, I'm going to cheat and just do stuff in small increments, since it's only 1 object moving here
            const float ballIncAmount=0.01f;
            float ballTimeLeft=time;
            while (ballTimeLeft>0.0f)
            {
                float adv=ballTimeLeft;
                if (adv>ballIncAmount) adv=ballIncAmount;
                ballTimeLeft-=adv;

                //now update it by adv and check collections and stuff
                if (fabs(ballDir.x())<0.5f)
                    ballDir.x()+=(rndf()<0?-rndf():rndf())*adv;

                ball+=ballDir*adv;

                //TEMP: control the ball for testing:
                /*if (INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::HOME))
                    ball.y()+=adv;
                if (INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::END))
                    ball.y()-=adv;
                if (INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::DELETE))
                    ball.x()-=adv;
                if (INPUT::KEYBOARD::IsKeyDown(INPUT::KEYBOARD::PAGE_DOWN))
                    ball.x()+=adv;*/

                //borders
                if (ball.y()>wallDistance)
                {
                    ball.y()=wallDistance-0.0001f;
                    ballDir.y()*=-1;
                }

                if (ball.y()<-wallDistance)
                {
                    ball.y()=-wallDistance+0.0001f;
                    ballDir.y()*=-1;
                }

                //paddle (hacky work, not technically accurate)
                //for this we use a point in the direction of the ball from the paddle that roughly follows the shape of the paddle to detect
                {
                    GEO::Vector2 paddleLeftPos(-paddleX, paddleY);
                    GEO::Vector2 reachToBall=(ball-paddleLeftPos).Normal() * Renderer::paddleLeft.radius * 1.3f;
                    reachToBall.x()/=Renderer::paddleLeft.ymultiplier;
                    reachToBall.y()*=2;
                    GEO::Vector2 reachPoint=paddleLeftPos+reachToBall;
                    if (GEO::VecDistance(reachPoint, ball)<Renderer::ball.radius+0.2f)
                    {
                        GEO::Vector2 effectiveNormal=reachToBall;
                        effectiveNormal.x()*=10;

                        if (GEO::VecDot(ballDir.Normal(), (ball-paddleLeftPos).Normal())<0) //different directions, do reflect
                            ballDir=GEO::VecReflect(ballDir, effectiveNormal.Normal());
                        else //same direction.. push it
                            ballDir+=(ball-paddleLeftPos)*adv;

                        reachPoint=paddleLeftPos+reachToBall*1.1f;
                        while (GEO::VecDistance(reachPoint, ball)<Renderer::ball.radius)
                            ball+=ballDir*0.0001f;
                    }
                }
                {
                    GEO::Vector2 paddleRightPos(paddleX, paddleY);
                    GEO::Vector2 reachToBall=(ball-paddleRightPos).Normal() * Renderer::paddleRight.radius * 1.3f;
                    reachToBall.y()*=2;
                    reachToBall.x()/=Renderer::paddleRight.ymultiplier;
                    GEO::Vector2 reachPoint=paddleRightPos+reachToBall;
                    if (GEO::VecDistance(reachPoint, ball)<Renderer::ball.radius+0.2f)
                    {
                        GEO::Vector2 effectiveNormal=reachToBall;
                        effectiveNormal.x()*=10;

                        if (GEO::VecDot(ballDir.Normal(), (ball-paddleRightPos).Normal())<0) //different directions, do reflect
                            ballDir=GEO::VecReflect(ballDir, effectiveNormal.Normal());
                        else //same direction.. push it
                            ballDir+=(ball-paddleRightPos)*adv;

                        reachPoint=paddleRightPos+reachToBall*1.1f;
                        while (GEO::VecDistance(reachPoint, ball)<Renderer::ball.radius)
                            ball+=ballDir*0.0001f;
                    }
                }

                //out of bounds
                if (ball.x()<-outOfBoundDistance || ball.x()>outOfBoundDistance)
                {
                    stateTime=0;
                    state=Ending;
                }
            }
        }
        else if (state==Ending) //ball went off the edge
        {
            if (stateTime>1.0f)
            {
                stateTime=0;
                state=Starting;
                paddleYResetMagnitude=-paddleY/2;
            }
            else
            {
                Renderer::ball.radius=(1-stateTime)*Renderer::ballSize;
            }
        }

        //sync renderer to state
        Renderer::paddleLeft.pos.y()=paddleY;
        Renderer::paddleLeft.pos.x()=-paddleX;
        Renderer::paddleRight.pos.y()=paddleY;
        Renderer::paddleRight.pos.x()=paddleX;
        Renderer::ball.pos.x()=ball.x();
        Renderer::ball.pos.y()=ball.y();
    }
}
