//The ray tracer.
//Luke Lenhart, 2008
//See the MPMA Framework license for details on how this code may be used.

#include "mpma/base/ThreadedTask.h"
#include "mpma/base/Memory.h"
#include "mpma/base/Profiler.h"
#include "mpma/gfxsetup/Setup.h"
#include "mpma/gfxsetup/GL.h"
#include "mpma/geo/Geo.h"
#include "mpma/geo/GeoObjects.h"
#include "mpma/geo/GeoIntersect.h"

#ifdef _DEBUG
    #include "mpma/input/Mouse.h" //for click-on-pixel to break debugging
    #include "mpma/base/Debug.h"
#endif

//define this to turn off shadows
//#define DISABLE_SHADOWS

//public data
namespace Renderer
{
    GEO::Sphere ball(GEO::Vector3(0, 0, 15), 0.5f);
    GEO::Vector3 lightPos(0, 0, -10);
    GEO::XAlignedUnboundedCylinder borderTop(6, 15, 1);
    GEO::XAlignedUnboundedCylinder borderBot(-6, 15, 1);
    GEO::YAlignedEllipsoid paddleLeft(GEO::Vector3(-7, 0, 15), 0.333f, 3);
    GEO::YAlignedEllipsoid paddleRight(GEO::Vector3(7, 0, 15), 0.333f, 3);
}
using namespace Renderer;

//internals
namespace
{
    //dimensions of window
    uint windowHeight=0;
    uint windowHeightHalf=0;
    uint windowWidth=0;
    uint windowWidthHalf=0;

    //buffer to write pixel colors to
    uint pixelBufferSize=0;
    float *pixelBuffer=0;
    int pixelJitter=0;

    //static scene setup
    const float lightMagnitude=400.0f;
    const float lightAmbient=0.15f;
    const float lightSpecular=0.1f;

    const GEO::Plane bgPlane(GEO::Vector3(0, 0, 40), GEO::Vector3(0, 0, -1));
    //const GEO::AARectoid borderTop(GEO::Vector3(0, 5,  20), 5, 1, 1);
    //const GEO::AARectoid borderBot(GEO::Vector3(0, -5, 20), 5, 1, 1);

    //calculates the visible color of an object given intersection information
    GEO::Vector3 CalculateReflectionColor(const GEO::Vector3 &point, const GEO::Vector3 &normal, const GEO::Vector3 &color)
    {
#ifdef _DEBUG
        if (fabs(normal.Length()-1)>0.0001f)
            throw "normal is not normalized.";
#endif

        //calc some initial stuff
        GEO::Vector3 adjustedIsectPoint=point+normal*0.01f;

        GEO::Vector3 isectToLight=lightPos-point;
        isectToLight.Normalize();
        GEO::Line rayIsectToLight(adjustedIsectPoint, isectToLight);

        float distToLight=GEO::VecDistance(point, lightPos);
        float attenuationModifier=lightMagnitude/(distToLight*distToLight);
        GEO::Vector3 shadowColor=color*lightAmbient*attenuationModifier;

#ifndef DISABLE_SHADOWS
        //see if its in shadow of the ball
        if (GEO::INTERSECT::RaySphere(rayIsectToLight, ball))
            return shadowColor;

        //see if it's in shadow of the top or bottom border
        if (point.y()>2 && GEO::INTERSECT::RayXAlignedUnboundedCylinder(rayIsectToLight, borderTop))
            return shadowColor;
        if (point.y()<-2 && GEO::INTERSECT::RayXAlignedUnboundedCylinder(rayIsectToLight, borderBot))
            return shadowColor;

        //see if it's in shadow of a paddle
        if (point.x()<0 && GEO::INTERSECT::RayYAlignedEllipsoid(rayIsectToLight, paddleLeft))
            return shadowColor;
        if (point.x()>0 && GEO::INTERSECT::RayYAlignedEllipsoid(rayIsectToLight, paddleRight))
            return shadowColor;
#endif

        //not in shadow, so return regular calc
        float diffuseModifier=GEO::VecDot(isectToLight.Normal(), normal);
        if (diffuseModifier<0) diffuseModifier=0;
        diffuseModifier+=lightAmbient;

        float specularAddition=GEO::VecDot(point.Normal(), GEO::VecReflect(isectToLight, normal).Normal());
        if (specularAddition<0.00001f) specularAddition=0.00001f;
        specularAddition*=specularAddition*specularAddition;

        return color*diffuseModifier*attenuationModifier+specularAddition*lightSpecular;
    }

    //calculates the color of a pixel by casting a ray that starts from 0,0
    GEO::Vector3 CalculatePixelColor(const GEO::Vector3 &rayDirection)
    {
        GEO::Line rayLine(GEO::Vector3(0, 0, 0), rayDirection);

        //see if it hit the ball
        GEO::Vector3 isect0, isect1;
        if (GEO::INTERSECT::LineSphere(rayLine, ball, isect0, isect1))
        {
            GEO::Vector3 &nearIsect=(isect0.z()<isect1.z()?isect0:isect1);
            GEO::Vector3 ballNorm=(nearIsect-ball.pos).Normal();
            return CalculateReflectionColor(nearIsect, ballNorm, GEO::Vector3(0.7f, 1.0f, 0.85f));
        }

        //see if it hit the top or bottom border
        /*GEO::Vector3 isectNorm0, isectNorm1;
        if (GEO::INTERSECT::LineAARectoid(rayLine, borderTop, isect0, isectNorm0, isect1, isectNorm1))
        {
            GEO::Vector3 &nearIsect=(isect0.z()<isect1.z()?isect0:isect1);
            GEO::Vector3 &nearNorm=(isect0.z()<isect1.z()?isectNorm0:isectNorm1);
            return CalculateReflectionColor(nearIsect, nearNorm, 0.8f);
        }

        if (GEO::INTERSECT::LineAARectoid(rayLine, borderBot, isect0, isectNorm0, isect1, isectNorm1))
        {
            GEO::Vector3 &nearIsect=(isect0.z()<isect1.z()?isect0:isect1);
            GEO::Vector3 &nearNorm=(isect0.z()<isect1.z()?isectNorm0:isectNorm1);
            return CalculateReflectionColor(nearIsect, nearNorm, 0.8f);
        }*/

        if (rayDirection.y()>0.28f && GEO::INTERSECT::LineXAlignedUnboundedCylinder(rayLine, borderTop, isect0, isect1))
        {
            GEO::Vector3 &nearIsect=(isect0.z()<isect1.z()?isect0:isect1);
            GEO::Vector3 norm=GEO::INTERSECT::NormalFromIntersection(nearIsect, borderTop);
            return CalculateReflectionColor(nearIsect, norm, GEO::Vector3(0.7f, 0.8f, 1.0f));
        }

        if (rayDirection.y()<-0.28f && GEO::INTERSECT::LineXAlignedUnboundedCylinder(rayLine, borderBot, isect0, isect1))
        {
            GEO::Vector3 &nearIsect=(isect0.z()<isect1.z()?isect0:isect1);
            GEO::Vector3 norm=GEO::INTERSECT::NormalFromIntersection(nearIsect, borderBot);
            return CalculateReflectionColor(nearIsect, norm, GEO::Vector3(0.7f, 0.8f, 1.0f));
        }

        //see if it hit a paddle
        if (rayDirection.x()<-0.39f && GEO::INTERSECT::LineYAlignedEllipsoid(rayLine, paddleLeft, isect0, isect1))
        {
            GEO::Vector3 &nearIsect=(isect0.z()<isect1.z()?isect0:isect1);
            GEO::Vector3 ballNorm=(nearIsect-paddleLeft.pos).Normal();
            return CalculateReflectionColor(nearIsect, ballNorm, GEO::Vector3(1.0f, 0.8f, 0.8f));
        }

        if (rayDirection.x()>0.39f && GEO::INTERSECT::LineYAlignedEllipsoid(rayLine, paddleRight, isect0, isect1))
        {
            GEO::Vector3 &nearIsect=(isect0.z()<isect1.z()?isect0:isect1);
            GEO::Vector3 ballNorm=(nearIsect-paddleRight.pos).Normal();
            return CalculateReflectionColor(nearIsect, ballNorm, GEO::Vector3(1.0f, 0.8f, 0.8f));
        }

        //by default we hit the background plane since we hit nothing else
        GEO::INTERSECT::LinePlane(rayLine, bgPlane, isect0);
        return CalculateReflectionColor(isect0, bgPlane.Normal(), 0.5f);
    }

    //processes each pixel
    void DrawRunner(uint pixelNumber, int unused)
    {
        //calculate the ray into the scene
        uint xPixel=pixelNumber%windowWidth;
        uint yPixel=pixelNumber/windowWidth;

        //in debug mode we can click on a pixel to break there
#ifdef _DEBUG
        if (INPUT::MOUSE::IsButtonDown(INPUT::MOUSE::LEFT_BUTTON))
        {
            INPUT::MOUSE::MousePixelPosition mp=INPUT::MOUSE::GetPixelPosition();
            if (mp.X==xPixel && mp.Y==yPixel)
                BREAKPOINT;
        }
#endif

        GEO::Vector3 direction(((int)xPixel-(int)windowWidthHalf)/(float)windowWidth, ((int)yPixel-(int)windowHeightHalf)/(float)windowHeight, 1);
        direction.Normalize();

        //do ray trace to get color
        float *pixels=&pixelBuffer[pixelNumber*3];
        *(GEO::Vector3*)pixels=CalculatePixelColor(direction);

        //apply super-simple ghetto dithering
        if ((pixelJitter+pixelNumber)&1)
            *(GEO::Vector3*)pixels+=GEO::Vector3(1/100.0f, 1/100.0f, 1/100.0f);
    }
}

//interface implementation
namespace Renderer
{
    //Draws the entire scene
    void DrawScene()
    {
        const GFX::GraphicsSetup *window=GFX::GetWindowState();
        windowWidth=window->Width;
        windowWidthHalf=windowWidth/2;
        windowHeight=window->Height;
        windowHeightHalf=windowHeight/2;

        //size gl viewport to match
        glViewport(0, 0, windowWidth, windowHeight);

        //realloc the pixel buffer if needed
        if (windowWidth*windowHeight!=pixelBufferSize)
        {
            pixelBufferSize=windowWidth*windowHeight*3;
            if (pixelBuffer)
                delete3_array(pixelBuffer);
            pixelBuffer=new3_array(float,pixelBufferSize);
        }

        //do every pixel
        pixelJitter=(pixelJitter+1)%2;

        ProfileStart("Ray Rendering");
        MPMA::ExecuteThreadedTask<void(*)(uint,int),DrawRunner>(windowWidth*windowHeight, 0);
        ProfileStop("Ray Rendering");

        //copy pixels to the framebuffer
        ProfileStart("Blit Image");
        glDrawPixels(windowWidth, windowHeight, GL_RGB, GL_FLOAT, pixelBuffer);
        ProfileStop("Blit Image");
    }

    //frees any resources the renderer allocated
    void Cleanup()
    {
        if (pixelBuffer)
            delete3_array(pixelBuffer);
        pixelBuffer=0;
    }
}
