//!\file Locks.h Thread locking constructs and atomic functions.
//Luke Lenhart, 2007
//See /docs/License.txt for details on how this code may be used.

#include "Types.h"

#ifndef LOCKS_H_INCLUDED
#define LOCKS_H_INCLUDED

namespace MPMA
{

    //  -- Atomic Operations safe for multiprocessor operations

    //!Atomically increments an integer.
    void AtomicIntInc(volatile uint *pint);

    //!Atomically decrements an integer.
    void AtomicIntDec(volatile uint *pint);

    //!Atomically adds one integer to another and returns the value of the original.
    int AtomicIntAdd(volatile int *pint, int addValue);

    //!Compares expectedValue with the value at pInt, and if they are the same, sets pInt to newValue and returns true with outResultValue set to newValue.  If they are different then pInt is unaffected, and returns false with outResultValue set to the value that was found at pInt.
    bool AtomicCompareExchange(volatile uint *pInt, uint expectedValue, uint newValue, volatile uint &outResultValue);


    // -- Locking constructs

    //!\brief ReaderWriter sleep-lock (re-entrant safe).
    //!Read locks don't block each other out, but Write locks are exclusive to all other locks.
    class RWSleepLock
    {
        friend class TakeRWSleepLock;

    public:
        RWSleepLock();
        ~RWSleepLock();

    protected:
        uint cntRead, cntWrite; //lock count for read/write access
        uint cntWritePend, cntReadPend; //pending access counter

        //you cannot duplicate this
        RWSleepLock(const RWSleepLock&);
        const RWSleepLock& operator=(const RWSleepLock&);
    };

    //!Used to lock/unlock a RWSleepLock automatically using scope.  The lock is released on dustruction.
    class TakeRWSleepLock
    {
    public:
        //!ctor - takes the lock (optionally), and releases it upon destruction
        inline TakeRWSleepLock(RWSleepLock &rwsSection, bool writeAccess=true, bool takeNow=true): crit((RWSleepLock&)rwsSection), meIn(false)
            {if (takeNow) Take(writeAccess); }
        inline ~TakeRWSleepLock()
            {  if (meIn) Leave(); }

        //!Manually takes the lock. (safe to call repeatedly)
        void Take(bool writeAccess);
        //!Manually releases the lock. (safe to call repeatedly)
        void Leave();

    private:
        mutable RWSleepLock &crit;

        bool meIn;
        bool meWrite;

        //you cannot duplicate this
        TakeRWSleepLock(const TakeRWSleepLock&);
        const TakeRWSleepLock& operator=(const TakeRWSleepLock&);
    };


    //!Mutex (re-entrant safe).
    class MutexLock
    {
        friend class TakeMutexLock;

    public:
        MutexLock();
        ~MutexLock();

    protected:
        void *crit;
        bool isMemManaged; //special case, since memory manager uses us

        //you cannot duplicate this
        MutexLock(const MutexLock&);
        const MutexLock& operator=(const MutexLock&);
    };

    //!Used to lock/unlock a MutexLock automatically using scope.  The lock is released on dustruction.
    class TakeMutexLock
    {
    public:
        //!ctor - takes the lock and releases it upon destruction
        TakeMutexLock(MutexLock &critSection);
        ~TakeMutexLock();

    private:
        MutexLock &crit;

        //you cannot duplicate this
        TakeMutexLock(const TakeMutexLock&);
        const TakeMutexLock& operator=(const TakeMutexLock&);
    };


    //!Spinlock (NOT re-entrant safe) that reverts to a sleeplock on single cpu systems.  If no lock-contention occurs, these is extremely little overhead in use.
    class SpinLock
    {
        friend class TakeSpinLock;
    protected:
        volatile uint taken;
        bool collision;
    public:
        inline SpinLock(): taken(0), collision(false) {}

        //you cannot duplicate this
        SpinLock(const SpinLock&);
        const SpinLock& operator=(const SpinLock&);
    };

    //!Used to lock/unlock a spinlock automatically using scope.  The lock is released on destruction.
    class TakeSpinLock
    {
    public:
        //!ctor - takes the lock (optionally), and releases it upon destruction
        inline TakeSpinLock(SpinLock &slock, bool takeNow=true): locker(slock), taken(false)
            { if (takeNow) Take(); }
        inline ~TakeSpinLock()
            { if (taken) Leave(); }

        //!Manually takes the lock. (NOT safe to call repeatedly)
        inline void Take();
        //!Manually releases the lock. (NOT safe to call repeatedly)
        inline void Leave();

    private:
        SpinLock &locker;
        bool taken;

        //you cannot duplicate this
        TakeSpinLock(const TakeSpinLock&);
        const TakeSpinLock& operator=(const TakeSpinLock&);
    };


    // -- Signal-based blocking constructs

    //!An object that can be used to block a thread until another thread signals it to resume.
    class BlockingObject
    {
    public:
        BlockingObject();
        ~BlockingObject();

        //!Marks the object as cleared, so nothing is blocking on it.
        void Clear();

        //!Marks the object as set, so BlockUntilClear will block on it.
        void Set();

        //!Blocks and waits for as long as the object is set (or if time is non-max, then until that many milliseconds pass).  Returns false only on timeout.  If setOnReturn is true, the object will be Set() when the function returns.  Only 1 thread should block on this, though any number may set or clear it.
        bool WaitUntilClear(bool setOnReturn=false, uint timeToWait=0xffffffff);

    private:
        volatile int clearCount;
        volatile bool isSet;
        void *pdata; //platform-specific container

        //you cannot duplicate this
        BlockingObject(const BlockingObject&);
        const BlockingObject& operator=(const BlockingObject&);
    };

} //namespace MPMA

//include the templated and inline code
#ifndef LOCKS_INCLUDE_INLINE
    #define LOCKS_INCLUDE_INLINE
    #include "Locks.cpp"
#endif

//include the platform-specific headers
#ifdef WIN32
    #include "win32/Locks.h"
#else
    #include "linux/Locks.h"
#endif

#endif //LOCKS_H_INCLUDED
